import atexit
import importlib
import multiprocessing.connection
from multiprocessing import Pipe
from multiprocessing.managers import DictProxy
from types import NoneType
from typing import Any, Callable

from loguru import logger

from more_than_inference.utils.fix_bug39959 import start_a_forkserver_process
from more_than_inference.utils.logging import *
from more_than_inference.utils.parser import CLI
from more_than_inference.utils.register import META


def add_runtime_path():
    import os
    import sys
    current_script_path = os.path.abspath('')
    if current_script_path not in sys.path:
        sys.path.append(current_script_path)


@META.regist_runner("ASYNC_OD")  # Async ObjectDetection
def runner_object_detection_async(args: CLI, inference_ally: DictProxy, pipe: multiprocessing.connection.Connection, unknown_args: Any = None):
    add_runtime_path()
    from more_than_inference.inference.object_detection import ObjectDetectionBase
    module = importlib.import_module(inference_ally["custom_class_module"])
    object_detection_class: type[ObjectDetectionBase] = getattr(module, inference_ally["custom_class_name"])
    object_detection_instance: ObjectDetectionBase = object_detection_class()
    object_detection_instance.store_args(args=args, unknown_args=unknown_args)
    object_detection_instance.load_model()
    atexit.register(object_detection_instance.unload_model)
    runner_name = getattr(runner_object_detection_async, "__runner_name__", "UNREGISTED_RUNNER")
    while True:
        logger.info(f"MTI::{runner_name}| WAITING ASYNC")
        try:
            object_detection_instance.run_async(pipe)
        except Exception:
            logger.exception(f"MTI::{runner_name}| EXCEPTION")
        logger.info(f"MTI::{runner_name}| INFER OVER")


@META.regist_runner("ASYNC_IG")  # Async ImageGeneration
def runner_image_generation_async(args: CLI, inference_ally: DictProxy, pipe: multiprocessing.connection.Connection, unknown_args: Any = None):
    add_runtime_path()
    from more_than_inference.inference.image_generation import ImageGenerationBase
    module = importlib.import_module(inference_ally["custom_class_module"])
    object_detection_class: type[ImageGenerationBase] = getattr(module, inference_ally["custom_class_name"])
    object_detection_instance: ImageGenerationBase = object_detection_class()
    object_detection_instance.store_args(args=args, unknown_args=unknown_args)
    object_detection_instance.load_model()
    atexit.register(object_detection_instance.unload_model)
    runner_name = getattr(runner_image_generation_async, "__runner_name__", "UNREGISTED_RUNNER")
    while True:
        logger.info(f"MTI::{runner_name}| WAITING ASYNC")
        try:
            object_detection_instance.run_async(pipe)
        except Exception:
            logger.exception(f"MTI::{runner_name}| EXCEPTION")
        logger.info(f"MTI::{runner_name}| INFER OVER")


@META.regist_runner("ASYNC_HTTP")
def runner_async_http(args: CLI, inference_ally: DictProxy, pipe: multiprocessing.connection.Connection, unknown_args: Any = None):
    from more_than_inference.extensions.fastapi.server import Server
    server = Server(args.port)
    runner_name = getattr(runner_async_http, "__runner_name__", "UNREGISTED_RUNNER")
    logger.info(f"MTI::{runner_name}| HTTP SERVICE START")
    get_callback = getattr(server, META.servers[args.runner])
    server_name = getattr(get_callback, "__server_name__", "UNREGISTED_SERVER")
    logger.info(f"MTI::{server_name}| SERVER LOADED")
    callback = get_callback(pipe)
    server.run(callback)


def run_async(args: CLI, inference_ally: DictProxy, unknown_args=None):
    assert args.runner in META.runners, f"undefined runner `{args.runner}`"
    p1, p2 = Pipe()
    algo_runtime = start_a_forkserver_process(target=META.runners[args.runner], args=(args, inference_ally, p1, unknown_args))
    algo_runtime.start()
    server_runtime_name = f"ASYNC_{args.service}"
    logger.info(f"finding `{server_runtime_name}` in registed runners.")
    try:
        server_runtime_func: Callable[[CLI, DictProxy, multiprocessing.connection.Connection, Any],
                                      NoneType] = META.runners[server_runtime_name]
        logger.info(f"found `{server_runtime_name}` in registed runners.")
        server_runtime_func(args, inference_ally, p2, unknown_args)
    except KeyboardInterrupt as e:
        logger.info(f"EXIT WITH KeyboardInterrupt")
    except Exception as e:
        logger.error(f"error in trying execute runner `{server_runtime_name}` for reason: {e}, stack trace:")
        logger.exception(e)
    finally:
        algo_runtime.kill()
        algo_runtime.join()
    logger.info(f"OVER.")


def run_pipeline(args: CLI, inference_ally: DictProxy, unknown_args: Any = None):
    ...


def run(args: CLI, inference_ally, unknown_args: Any = None):
    if args.runner.startswith("ASYNC"):
        run_async(args, inference_ally, unknown_args)
    elif args.runner.startswith("PIPE"):
        run_pipeline(args, inference_ally, unknown_args)
    else:
        raise ValueError(f"unsupport runner mode `{args.runner}`")



import os
from pathlib import Path

from more_than_inference.utils.register import META


def build_proto(dist_dir, proto_dir):  # install mypy-protobuf first
    os.system(f'protoc --python_out={dist_dir} --mypy_out={dist_dir} --proto_path={dist_dir} {proto_dir}/*.proto')


def clean_proto(proto_dir):
    os.system(f'rm -fr {proto_dir}/*_pb2.py')


@META.regist_cmd()
def refresh_proto(args):
    dist_root = Path(__file__).resolve().parent.parent
    dist_dir = dist_root.parent
    proto_dir = dist_root / "protos"
    clean_proto(proto_dir)
    build_proto(dist_dir, proto_dir)



import importlib
import importlib.util
import sys
from pathlib import Path
from typing import Union


def get_module_name(relative_path: Union[Path, str]):
    path = Path(relative_path).resolve()
    cwd = Path.cwd()
    try:
        relative_path = path.relative_to(cwd)
        module_name = str(relative_path).replace('/', '.').replace('\\', '.')
        if path.is_file():
            module_name = module_name.rsplit('.', 1)[0]
        return True, module_name
    except ValueError:
        return False, "Path is not a subpath of the current directory."


def load_module_by_spec(main_script: Union[Path, str]):
    flag, ret = get_module_name(main_script)
    if not flag:
        return False, ret
    spec = importlib.util.spec_from_file_location(ret, main_script)
    main = importlib.util.module_from_spec(spec)
    sys.modules[ret] = main
    spec.loader.exec_module(main)
    return True, "module loaded"



from multiprocessing import Manager, Process
from multiprocessing.managers import DictProxy

from more_than_inference.runs import *
from more_than_inference.utils.build_proto import *
from more_than_inference.utils.module_functions import load_module_by_spec
from more_than_inference.utils.parser import CLI, apply_act, parse_act, parse_runner, parse_script, parse_service
from more_than_inference.utils.register import META


def run_cmd(args: CLI):
    if args.script in META.cmds:
        try:
            META.cmds[args.script](args)
        except:
            logger.exception(f"cmd `{args.script}` exec with error")
        finally:
            return True
    return False


def test_import(args: CLI, inference_ally: DictProxy):
    flag, msg = load_module_by_spec(args.script)
    logger.info(msg)
    if not flag:
        inference_ally["custom_class_flag"] = False
        return
    meta_class = META.registers[args.type]
    meta_class.print_inferences()
    custom_class = meta_class.get_inference(args.name, args.module)
    if custom_class is not None:
        inference_ally["custom_class_flag"] = True
        inference_ally["custom_class_name"] = custom_class.__name__
        inference_ally["custom_class_module"] = custom_class.__module__
    else:
        inference_ally["custom_class_flag"] = False


def run_script(args: CLI, unknown_args):
    manager = Manager()
    inference_ally: DictProxy = manager.dict()
    ally = Process(target=test_import, args=(args, inference_ally))
    ally.start()
    ally.join()
    logger.info(f"run_script:: inference_ally: {inference_ally}")
    if inference_ally["custom_class_flag"]:
        logger.info(f"run_script:: custom_class_name: {inference_ally['custom_class_name']}")
    else:
        raise ValueError(f"Model type '{args.type}' with name '{args.name}' not found.")

    run(args, inference_ally, unknown_args)


@META.regist_cmd()
def types(args): print("Support task types:", list(META.cmds.keys()))


def main():
    parser = parse_script()
    parser = parse_runner(parser)
    parser = parse_service(parser)
    parser = parse_act(parser)

    args, unknown_args = parser.parse_known_args()
    logger.info(f"cmd `{args.script}` args `{args}`, unknown_args `{unknown_args}` start")
    args: CLI
    # reconfig act args
    args = apply_act(args, unknown_args)

    if run_cmd(args):
        logger.info(f"cmd `{args.script}` args `{args}` over")
    elif args.script.endswith(".py"):
        run_script(args, unknown_args)
    else:
        raise ValueError(f"not support command `{args.script}` current support is: {list(META.cmds.keys())}")


if __name__ == "__main__":
    main()



import multiprocessing.connection
import os
from multiprocessing import Queue, shared_memory
from typing import Any

import numpy as np
from loguru import logger

from more_than_inference.datas.object_detection import ObjectDetectionInputSerializer, ObjectDetectionOutputSerializer
from more_than_inference.extensions.zmq.cs_base import ZeroMQClient, ZeroMQServer
from more_than_inference.utils.serializer import TypedNode, TypedPipeline


class ObjectDetectionServer(ZeroMQServer):
    def __init__(self, bind_addr: str, node_name: str, node_key: str, domain: str, node_category: str):
        super().__init__(bind_addr, node_name, node_key, domain)
        self.node_category = node_category

    def run_async(self, pipe: multiprocessing.connection.Connection,
                  typed_input: type[ObjectDetectionInputSerializer],
                  typed_output: type[ObjectDetectionOutputSerializer]):
        """`processing_async` run outside the inference process
        """
        def _get_input(_d: bytes):
            _inst = typed_input()
            _inst.loads(_d)
            return _inst

        def _get_output(output: Any) -> ObjectDetectionOutputSerializer:
            _inst = typed_output()
            _inst.result = output
            return _inst

        def processing_async(data_in: bytes) -> bytes:
            _input = _get_input(data_in)
            _image = _input.image
            if _image is None:
                return b'E: Image Decode Error.'
            shm = shared_memory.SharedMemory(create=True, size=_image.nbytes)
            shared_arr = np.ndarray(_image.shape, dtype=_image.dtype, buffer=shm.buf)
            np.copyto(shared_arr, _image)
            pipe.send((shm.name, _image.shape, _image.dtype))
            result = pipe.recv()
            shm.close()
            shm.unlink()
            output = _get_output(result)
            return output.dumps()
        return processing_async

    def run_pipeline(self, q_in: Queue, q_out: Queue):
        """`processing_input` run outside the inference process,
        `processing_output` run outside the inference process.
        """

        def _get_output(output: Any) -> TypedPipeline:
            _inst = TypedPipeline()
            _inst.result = output
            return _inst

        def processing_input(data_in: bytes) -> bytes:
            q_in.put(data_in)
            return b'I: OK'

        def processing_output(client: ObjectDetectionClient):
            while True:
                tp: TypedPipeline = q_out.get()
                if len(tp.StackList) == 0:
                    logger.info(f"Pipeline Ended, Key is {tp.Key}.")
                next_node = tp.StackList.pop()
                client.send_node(next_node, tp.dumps())

        return processing_input, processing_output


class ObjectDetectionClient(ZeroMQClient):
    def ipc_exists(self, sock: str):
        sock = sock.replace("ipc://", "")
        return os.path.exists(sock)

    async def send_node(self, node: TypedNode, data: bytes):
        sock = node.LocalSocket
        if not self.ipc_exists(sock):
            sock = node.RemoteSocket

        self.socket.connect(sock)
        await self.socket.send(data)
        reply = await self.socket.recv()
        self.socket.disconnect(sock)
        logger.info(reply)
        return reply


if __name__ == "__main__":
    import asyncio
    from multiprocessing import Pipe, Process

    import cv2

    from more_than_inference.datas.object_detection import TypedObjectDetectionArrayInput, TypedObjectDetectionPbOutput

    async def main():

        def process_pipe(pip_in: multiprocessing.connection.Connection):
            while True:
                shm_name, shm_shape, shm_dtype = pip_in.recv()
                logger.info(f"P: RECV {shm_name,shm_shape,shm_dtype}")
                existing_shm = shared_memory.SharedMemory(name=shm_name, create=False)
                shared_arr = np.ndarray(shm_shape, dtype=shm_dtype, buffer=existing_shm.buf)
                cv2.imwrite(shm_name+".png", shared_arr)
                existing_shm.close()
                pip_in.send({"key": "id:1",
                            "result": [{"type": "rectangle",
                                        "points": [0, 0, 152, 153],
                                        "confidence": 0.698,
                                        "label": "default"}]})

        ods = ObjectDetectionServer(
            bind_addr="ipc:///tmp/zero.sock",
            node_name="",
            node_key="",
            domain="",
            node_category="",
        )
        pipe_parent, pipe_child = Pipe()
        p = Process(target=process_pipe, args=(pipe_child,))
        p.start()
        processing_async = ods.run_async(pipe_parent, TypedObjectDetectionArrayInput, TypedObjectDetectionPbOutput)
        asyncio.create_task(ods.start(processing_async))

        await asyncio.sleep(1)

        din = TypedObjectDetectionArrayInput()
        din.image = cv2.imread("assets/Lenna.png")
        din.Key = "Lenna"

        odc = ObjectDetectionClient(bind_addr="",
                                    node_name="",
                                    node_key="",
                                    domain=""
                                    )
        node = TypedNode()
        node.LocalSocket = "ipc:///tmp/zero.sock"
        node.Key = "client"
        reply = await odc.send_node(node, din.dumps())
        dou = TypedObjectDetectionPbOutput()
        dou.loads(reply)
        print(dou.Key, dou.Result)
        p.join()

    asyncio.run(main())



import base64
import builtins
import json
from typing import Any, List, Optional

import cv2
import numpy as np
from addict import Dict

from more_than_inference.protos.object_detection_pb2 import (ObjectDetectionArrayInput, ObjectDetectionBase64Input,
                                                             ObjectDetectionJsonOutput, ObjectDetectionPbOutput,
                                                             ObjectDetectionResult)
from more_than_inference.utils.serializer import BaseSerializer, ProtoSerializer


class ObjectDetectionInputSerializer(BaseSerializer):
    @property
    def image(self) -> Optional[np.ndarray]: raise NotImplementedError
    @image.setter
    def image(self, image: np.ndarray): raise NotImplementedError


class ObjectDetectionResultSerializer(BaseSerializer):
    @property
    def result(self) -> Optional[np.ndarray]: raise NotImplementedError
    @result.setter
    def result(self, value: Dict[str, Any]): raise NotImplementedError


class ObjectDetectionOutputSerializer(BaseSerializer):
    @property
    def result(self) -> Optional[np.ndarray]: raise NotImplementedError
    @result.setter
    def result(self, value: List[Dict[str, Any]]): raise NotImplementedError


class TypedObjectDetectionArrayInput(ObjectDetectionInputSerializer, metaclass=ProtoSerializer):
    _data_class = ObjectDetectionArrayInput
    _protected = ["Key", "ArrayData", "Width", "Height", "Channels"]

    Key: builtins.str
    ArrayData: builtins.bytes
    Width: builtins.int
    Height: builtins.int
    Channels: builtins.int

    @property
    def image(self) -> Optional[np.ndarray]:
        if self.ArrayData is not None:
            image = np.frombuffer(self.ArrayData, dtype=np.uint8)
            return image.reshape((self.Height, self.Width, self.Channels))
        return None

    @image.setter
    def image(self, image: np.ndarray):
        self.Height, self.Width, self.Channels = image.shape
        self.ArrayData = image.tobytes()


class TypedObjectDetectionBase64Input(ObjectDetectionInputSerializer, metaclass=ProtoSerializer):
    _data_class = ObjectDetectionBase64Input
    _protected = ["Key", "B64Data"]

    Key: builtins.str
    B64Data: builtins.str

    @property
    def image(self) -> Optional[np.ndarray]:
        if self.B64Data:
            image_data = base64.b64decode(self.B64Data)
            image = cv2.imdecode(np.frombuffer(image_data, np.uint8), cv2.IMREAD_COLOR)
            return image
        return None

    @image.setter
    def image(self, image: np.ndarray):
        _, buffer = cv2.imencode('.jpg', image)
        self.B64Data = base64.b64encode(buffer)


class TypedObjectDetectionJsonOutput(ObjectDetectionInputSerializer, metaclass=ProtoSerializer):
    _data_class = ObjectDetectionJsonOutput
    _protected = ["Key", "JsonData"]

    Key: builtins.str
    JsonData: builtins.str

    @property
    def image(self) -> Optional[np.ndarray]:
        json_data = json.loads(self.JsonData)
        image_b64 = json_data.get('image')
        if image_b64:
            image_data = base64.b64decode(image_b64)
            image = cv2.imdecode(np.frombuffer(image_data, np.uint8), cv2.IMREAD_COLOR)
            return image
        return None

    @image.setter
    def image(self, image: np.ndarray):
        _, buffer = cv2.imencode('.jpg', image)
        image_b64 = base64.b64encode(buffer).decode('utf-8')
        json_data = json.loads(self.JsonData)
        json_data['image'] = image_b64
        self.JsonData = json.dumps(json_data)


class TypedObjectDetectionResult(ObjectDetectionResultSerializer, metaclass=ProtoSerializer):
    _data_class = ObjectDetectionResult
    _registed = {"Result": list[builtins.float]}

    Type: builtins.str
    Points: list[builtins.float]
    Confidence: builtins.float
    Label: builtins.str

    @property
    def result(self) -> Optional[np.ndarray]:
        return {
            "type": getattr(self, "Type", None),
            "points": list(getattr(self, "Points", None)),
            "confidence": getattr(self, "Confidence", None),
            "label": getattr(self, "Label", None),
        }

    @result.setter
    def result(self, value: Dict[str, Any]):
        self.Type = value.get("type", None)
        if "points" in value:
            self.Points = value.get("points")
        self.Confidence = value.get("confidence", None)
        self.Label = value.get("label", None)


class TypedObjectDetectionPbOutput(ObjectDetectionOutputSerializer, metaclass=ProtoSerializer):
    _data_class = ObjectDetectionPbOutput
    _registed = {"Result": list[TypedObjectDetectionResult]}

    Key: builtins.str
    Result: list[TypedObjectDetectionResult]

    @property
    def result(self) -> Optional[np.ndarray]:
        return {
            "key": getattr(self, "Key", None),
            "result": [i.result for i in getattr(self, "Result", [])]
        }

    @result.setter
    def result(self, value: List[Dict[str, Any]]):
        self.Key = value.get("key", None)
        self.Result = []
        for i in value.get("result", []):
            rr = TypedObjectDetectionResult()
            rr.result = i
            self.Result.append(rr)



import base64
import multiprocessing.connection
import tempfile
from multiprocessing import shared_memory
from pathlib import Path
from typing import Callable, Dict, List

import cv2
import numpy as np
from loguru import logger
from pydantic import BaseModel
from requests import Session

from more_than_inference.extensions.fastapi.server_base import FastServer
from more_than_inference.inference.image_generation import (ImageGenerationInput, ImageGenerationOutput, image_b64s_to_path,
                                                            image_path_to_b64s)
from more_than_inference.utils.register import META


class PredictRequest(BaseModel):
    images: List[str]
    image_type: str


class Server(FastServer):
    async def object_detection(self, images, image_type) -> Dict:
        imgs = []
        if image_type == 'url':
            for image in images:
                with Session() as client:
                    resp = await client.get(image)
                    byte_content = await resp.read()
                image = np.asarray(bytearray(byte_content), dtype="uint8")
                image = cv2.imdecode(image, cv2.IMREAD_COLOR)
                imgs.append(image)
        elif image_type == 'image':     # act前端 base64
            for image in images:
                image_data = base64.b64decode(image)
                image_array = np.frombuffer(image_data, np.uint8)
                image = cv2.imdecode(image_array, cv2.IMREAD_COLOR)
                imgs.append(image)
        else:
            raise ValueError(f'image_type is invalid!')

        result = self._callback_func(imgs)
        logger.info(f"RET: {result}")
        return result

    @META.regist_server("ASYNC_OD")
    def run_async_object_detection(self, pipe: multiprocessing.connection.Connection):
        def processing_async(imgs: list[np.ndarray]) -> dict:
            shms: list[shared_memory.SharedMemory] = []
            sends: List[tuple] = []
            for img in imgs:
                shm = shared_memory.SharedMemory(create=True, size=img.nbytes)
                shared_arr = np.ndarray(img.shape, dtype=img.dtype, buffer=shm.buf)
                np.copyto(shared_arr, img)
                shms.append(shm)
                sends.append((shm.name, img.shape, img.dtype))

            pipe.send(sends)
            result = pipe.recv()
            for shm in shms:
                shm.close()
                shm.unlink()
            return result
        return processing_async

    async def image_generation(self, ig_input: ImageGenerationInput) -> ImageGenerationOutput:
        ig_output: ImageGenerationOutput = self._callback_func(ig_input)
        logger.info(f"IG_STATUS: {ig_output.task_finished}")
        return ig_output

    @META.regist_server("ASYNC_IG")
    def run_async_image_generation(self, pipe: multiprocessing.connection.Connection):
        def processing_async(ig_input: ImageGenerationInput) -> ImageGenerationOutput:
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_dir_path = Path(temp_dir)
                base_path = temp_dir_path / "input"
                ig_input.input_images = image_b64s_to_path(ig_input.input_images, base_path)
                ig_input.output_directory = temp_dir_path
                pipe.send(ig_input)
                ig_output: ImageGenerationOutput = pipe.recv()
                ig_output.output_images = image_path_to_b64s(ig_output.output_images)
                return ig_output
        return processing_async

    def _setup_routes(self):
        super()._setup_routes()

        @self.app.post("/forward_ig")
        @self.app.get("/forward_ig")
        async def forward_ig(request: ImageGenerationInput):
            return await self.image_generation(request)

        @self.app.post("/predict_od")
        @self.app.get("/predict_od")
        async def predict_od(request: PredictRequest):
            return await self.object_detection(request.images, request.image_type)

        # add act service route
        @self.app.post("/autotable/predict")
        @self.app.get("/autotable/predict")
        async def act(request: PredictRequest):
            return await self.object_detection(request.images, request.image_type)

    def run(self, callback: Callable):
        self._callback_func = callback
        super().run()



import argparse
import sys
from dataclasses import dataclass, is_dataclass
from typing import Any, Dict, Type


def shallow_asdict(obj: Any) -> Dict[str, "ARG"]: return {field.name: getattr(obj, field.name)
                                                          for field in obj.__dataclass_fields__.values()} if is_dataclass(obj) else obj


@dataclass(frozen=True)
class ARG:
    short: str = None
    type: Type = str
    required: bool = False
    default: Any = None
    help: str = None

    def get_args(self, long: str): return [f"-{self.short}", f"--{long}"] if self.short else [f"--{long}"]
    def apply_args(self, val: Any): return [f"--{self.long}", val]


def parse_ln(ln: Dict[str, ARG], parser: argparse.ArgumentParser = None):
    if not parser:
        parser = argparse.ArgumentParser()
    for varg_name, varg in ln.items():
        parser.add_argument(*varg.get_args(long=varg_name), type=varg.type, required=varg.required,
                            default=varg.default, help=varg.help)
    return parser


@dataclass(frozen=True)
class RUNNER_CLI:
    type: ARG = ARG(short='t', type=str, default="ObjectDetection",
                    help='cli:: Model type, or task flag, default is `ObjectDetection`')
    name: ARG = ARG(short='n', type=str,
                    help='cli:: Model name, a specified name of the model, such as `CustomYOLO`, if not defined, cli whould use the first searched as default')
    module: ARG = ARG(short='m', type=str,
                      help='cli:: Model module, specify if name is used, help the cli locate the class, if not defined, behave as `cli:: name`.')
    runner: ARG = ARG(short='r', type=str, default="ASYNC_OD",
                      help='cli:: Runtime runner, default is `ASYNC_OD`(ASYNC_OD|ASYNC_IG), function defined in `more_than_inference/runs.py`, must start with ASYNC/PIPE')
    service: ARG = ARG(short='s', type=str, default="HTTP",
                       help='cli:: Runtime backend service, default is `HTTP`, enabled with extension installed')


@dataclass(frozen=True)
class SERVICE:
    port: ARG = ARG(type=int, default=3963, help='service:: service port')


@dataclass(frozen=True)
class ACT:
    act_model_path: ARG = ARG(type=str, help='act:: directory of the model')
    act_used_model: ARG = ARG(type=str, help='act:: name of the model(no ext)')
    act_service_port: ARG = ARG(type=int, help='act:: service port, if use, this will replace `service:: port`')


class CLI(RUNNER_CLI,
          SERVICE,
          ACT):
    script: str


def parse(): return argparse.ArgumentParser()
def parse_runner(parser: argparse.ArgumentParser = None): return parse_ln(shallow_asdict(RUNNER_CLI()), parser=parser)
def parse_service(parser: argparse.ArgumentParser = None): return parse_ln(shallow_asdict(SERVICE()), parser=parser)
def parse_act(parser: argparse.ArgumentParser = None): return parse_ln(shallow_asdict(ACT()), parser=parser)


def parse_script(parser: argparse.ArgumentParser = None):
    if not parser:
        parser = argparse.ArgumentParser()
    parser.add_argument(
        'script', help=f'Script to load models, or cli to do something, use `{sys.argv[0]} types` to list all cli')
    return parser


def apply_act(args: CLI, unknown_args: Any) -> CLI:
    parser_unknown = argparse.ArgumentParser()
    parser_unknown.add_argument("--service_port", type=int)
    unknown_args, _ = parser_unknown.parse_known_args(unknown_args)

    if args.act_service_port is None and unknown_args.service_port is None:
        return args
    setattr(args, "port", args.act_service_port or unknown_args.service_port)
    return args



import builtins
import multiprocessing.connection
from abc import abstractmethod
from pathlib import Path
from typing import Any, List, Union

from pydantic import BaseModel

class ImageGenerationInput(BaseModel):
    task_id: str
    models: str
    prompt: str
    negative_prompt: str
    input_images: List[Union[Path, str]]
    output_directory: Union[Path, str]
    steps: int
    seed: int
    generation_count: int

    def __init__(self, 
                 task_id: str = ..., 
                 models:str=...,
                 prompt: str = ..., 
                 negative_prompt: str = ..., 
                 input_images: List[Union[Path, str]] = ..., 
                 output_directory: Union[Path, str] = ..., 
                 steps: int = ..., 
                 seed: int = ..., 
                 generation_count: int = ...
                 ) -> None: ...

class ImageGenerationOutput(BaseModel):
    output_images: List[Union[Path, str]]
    trace_back_info: str
    task_finished: bool

    def __init__(self, 
                 output_images: List[Union[Path, str]] = ..., 
                 trace_back_info: str = ..., 
                 task_finished: bool = ...
                 ) -> None: ...

class ImageGenerationBase:
    def store_args(self:"ImageGenerationBase", args:Any=None, unknown_args:Any=None) -> None: ...
    @classmethod
    def print_inferences(cls:type["ImageGenerationBase"]): ...
    @classmethod
    def get_inference(cls:type["ImageGenerationBase"], name:builtins.str=None, module:builtins.str=None): ...
    @abstractmethod
    def load_model(self:"ImageGenerationBase", *args, **kwargs): ...
    @abstractmethod
    def unload_model(self:"ImageGenerationBase", *args, **kwargs): ...
    @abstractmethod
    def preprocess(self:"ImageGenerationBase", data: ImageGenerationInput, *args, **kwargs)->ImageGenerationInput: ...
    @abstractmethod
    def postprocess(self:"ImageGenerationBase", data: ImageGenerationOutput, *args, **kwargs)->ImageGenerationOutput: ...
    @abstractmethod
    def predict(self:"ImageGenerationBase", data: ImageGenerationInput, *args, **kwargs)->ImageGenerationOutput: ...
    def run(self:"ImageGenerationBase", data: Any, *args, **kwargs): ...
    def run_async(self: "ImageGenerationBase", pipe: multiprocessing.connection.Connection): ...
    def run_pipeline(self: "ImageGenerationBase", pipe: multiprocessing.connection.Connection): ...

def image_path_to_b64s(paths:Union[Path,List[Path]])->Union[str,List[str]]: ...
def image_b64s_to_path(strs:Union[str,List[str]], base:Path)->Union[Path,List[Path]]: ...


from typing import Callable


class METARegister:
    _instance = None
    registers = {}
    runners = {}
    cmds = {}
    servers = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def regist(self, name=None):
        def decorator(cls: type):
            nonlocal name
            if name is None:
                name = cls.__name__
            assert name not in self.registers, f"cls name: `{name}` already registed."
            self.registers[name] = cls
            if not hasattr(cls, 'sub_classes'):
                cls.sub_classes = []
            return cls
        return decorator

    def regist_runner(self, name=None):
        def decorator(func: Callable):
            nonlocal name
            if name is None:
                name = func.__name__
            assert name not in self.runners, f"func name: `{name}` already registed."
            self.runners[name] = func
            setattr(func, "__runner_name__", name)
            return func
        return decorator

    def regist_cmd(self):
        def decorator(func: Callable):
            name = func.__name__
            assert name not in self.cmds, f"cmd name: `{name}` already registed."
            self.cmds[name] = func
            setattr(func, "__runner_name__", name)
            return func
        return decorator

    def regist_server(self, name=None):
        def decorator(func: Callable):
            nonlocal name
            if name is None:
                name = func.__name__
            assert name not in self.servers, f"server name: `{name}` already registed."
            self.servers[name] = func.__name__
            setattr(func, "__server_name__", name)
            return func
        return decorator


META = METARegister()



from typing import Callable

import uvicorn
from fastapi import FastAPI
from loguru import logger


class FastServer:
    def __init__(self, port):
        self.port = port
        self.app = FastAPI()
        self._setup_routes()

    def _setup_routes(self):
        @self.app.get("/probe")
        @self.app.post("/probe")
        async def probe():
            """Health check."""
            return {"status": "OK"}

        @self.app.post("/start")
        @self.app.get("/start")
        async def start(message):
            return {"message": message}

    def run(self, callback: Callable = None):
        logger.info(f"HTTP_SERVICE LISTEN: `0.0.0.0:{self.port}`")
        if callback:
            self._callback_func = callback
        uvicorn.run(self.app, host="0.0.0.0", port=self.port, log_level="info", log_config=None)



import multiprocessing.connection
from multiprocessing import shared_memory
from typing import List

import numpy as np

from more_than_inference.inference.meta import ArgParserMixin, InferMixin, MetaInference, ModuleMixin, RuntimeMixin
from more_than_inference.utils.register import META


class RunBatchImagesAsyncMixin:
    def run_async(self: RuntimeMixin, pipe: multiprocessing.connection.Connection):
        kwdict, *pipe_inputs = pipe.recv()
        imgs = []
        for pipe_image in pipe_inputs:
            shm_name, shm_shape, shm_dtype = pipe_image
            existing_shm = shared_memory.SharedMemory(name=shm_name)
            imgs.append(np.ndarray(shm_shape, dtype=shm_dtype, buffer=existing_shm.buf))
            existing_shm.close()
        rsts = self.run(imgs, **kwdict)
        pipe.send(rsts)


@META.regist(name="Detection")
class DetectionBase(ArgParserMixin,
                    ModuleMixin,
                    InferMixin,
                    RuntimeMixin,
                    RunBatchImagesAsyncMixin,
                    metaclass=MetaInference):
    ...


def run_batch_async_detection(pipe: multiprocessing.connection.Connection):
    def processing_async(imgs: list[np.ndarray], kwdict: dict = None) -> dict:
        shms: list[shared_memory.SharedMemory] = []
        sends: List[tuple] = [kwdict or {}]
        for img in imgs:
            shm = shared_memory.SharedMemory(create=True, size=img.nbytes)
            shared_arr = np.ndarray(img.shape, dtype=img.dtype, buffer=shm.buf)
            np.copyto(shared_arr, img)
            shms.append(shm)
            sends.append((shm.name, img.shape, img.dtype))

        pipe.send(sends)
        result = pipe.recv()
        for shm in shms:
            shm.close()
            shm.unlink()
        return result
    return processing_async



import multiprocessing.connection
from multiprocessing import shared_memory

import numpy as np

from more_than_inference.inference.meta import ArgParserMixin, InferMixin, MetaInference, ModuleMixin, RuntimeMixin
from more_than_inference.utils.register import META


class RunAsyncMixin:
    def run_async(self: RuntimeMixin, pipe: multiprocessing.connection.Connection):
        pipe_imgs = pipe.recv()
        rsts = []
        for pipe_image in pipe_imgs:
            shm_name, shm_shape, shm_dtype = pipe_image
            existing_shm = shared_memory.SharedMemory(name=shm_name)
            rsts.append(self.run(np.ndarray(shm_shape, dtype=shm_dtype, buffer=existing_shm.buf)))
            existing_shm.close()
        pipe.send(rsts)


class RunPipelineMixin:
    def run_pipeline(self: RuntimeMixin, pipe: multiprocessing.connection.Connection): ...


@META.regist(name="ObjectDetection")
class ObjectDetectionBase(ArgParserMixin,
                          ModuleMixin,
                          InferMixin,
                          RuntimeMixin,
                          RunAsyncMixin,
                          RunPipelineMixin,
                          metaclass=MetaInference):
    ...



# This script is a fix for python issue 39959 (https://bugs.python.org/issue39959)
import sys
from multiprocessing import get_context
from multiprocessing.context import BaseContext
from multiprocessing.process import BaseProcess


def start_a_naive_process(group=None, target=None, name=None, args: tuple = None, kwargs: dict = None, daemon=None,
                          method: str = None,) -> BaseProcess:
    ctx: BaseContext = get_context(method)
    P: BaseProcess = ctx.Process
    return P(group=group, target=target, name=name, args=args or (), kwargs=kwargs or {}, daemon=daemon)


def start_a_spawn_process(group=None, target=None, name=None, args: tuple = None, kwargs: dict = None, daemon=None) -> BaseProcess:
    return start_a_naive_process(group, target, name, args, kwargs, daemon, "spawn")


def start_a_fork_process(group=None, target=None, name=None, args: tuple = None, kwargs: dict = None, daemon=None) -> BaseProcess:
    return start_a_naive_process(group, target, name, args, kwargs, daemon, "fork")


def start_a_forkserver_process(group=None, target=None, name=None, args: tuple = None, kwargs: dict = None, daemon=None) -> BaseProcess:
    return start_a_naive_process(group, target, name, args, kwargs, daemon, "forkserver")


if sys.platform == "win32":
    start_a_fork_process = start_a_spawn_process
    start_a_forkserver_process = start_a_spawn_process



import base64
import multiprocessing.connection
import uuid
from pathlib import Path
from typing import List, Union

from pydantic import BaseModel, Field

from more_than_inference.inference.meta import ArgParserMixin, InferMixin, MetaInference, ModuleMixin, RuntimeMixin
from more_than_inference.utils.register import META


class ImageGenerationInput(BaseModel):
    task_id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    models: str = "{}"
    prompt: str = ""
    negative_prompt: str = ""
    input_images: List[Union[Path, str]] = []
    output_directory: Union[Path, str] = None
    steps: int = 20
    seed: int = -1
    generation_count: int = 1


class ImageGenerationOutput(BaseModel):
    output_images: List[Union[Path, str]] = []
    trace_back_info: str = ""
    task_finished: bool = False


def image_path_to_b64s(paths: Union[Path, List[Path]]) -> Union[str, List[str]]:
    def path_to_b64(path: Path) -> str:
        with open(path, 'rb') as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    if isinstance(paths, list):
        return [path_to_b64(path) for path in paths]
    else:
        return path_to_b64(paths)


def image_b64s_to_path(strs: Union[str, List[str]], base: Path) -> Union[Path, List[Path]]:
    def b64_to_image(b64_str: str, file_path: Path):
        image_data = base64.b64decode(b64_str)
        with open(file_path, 'wb') as file:
            file.write(image_data)

    if isinstance(strs, list):
        paths = []
        for idx, b64_str in enumerate(strs):
            file_path = base.parent / f"{base.stem}_{idx+1}.png"
            b64_to_image(b64_str, file_path)
            paths.append(file_path)
        return paths
    else:
        file_path = base.with_suffix('.png')
        b64_to_image(strs, file_path)
        return file_path


class RunAsyncMixin:
    def run_async(self: RuntimeMixin, pipe: multiprocessing.connection.Connection):
        inputs: ImageGenerationInput = pipe.recv()
        outputs: ImageGenerationOutput = self.run(data=inputs)
        pipe.send(outputs)


class RunPipelineMixin:
    def run_pipeline(self: RuntimeMixin, pipe: multiprocessing.connection.Connection): ...


@META.regist(name="ImageGeneration")
class ImageGenerationBase(ArgParserMixin,
                          ModuleMixin,
                          InferMixin,
                          RuntimeMixin,
                          RunAsyncMixin,
                          RunPipelineMixin,
                          metaclass=MetaInference):
    ...



import argparse
import builtins

class RUNNER_CLI:
    type:builtins.str
    name:builtins.str 
    module:builtins.str 
    runner:builtins.str
    service:builtins.str

class SERVICE:
    port:builtins.int
class ACT:
    act_model_path:builtins.str 
    act_used_model:builtins.str
    act_service_port:builtins.int
class CLI:
    script:str

    type:builtins.str
    name:builtins.str 
    module:builtins.str 
    runner:builtins.str
    service:builtins.str
    
    port:builtins.int

    act_model_path:builtins.str
    act_used_model:builtins.str
    act_service_port:builtins.int

def parse()->argparse.ArgumentParser: ...
def parse_runner(parser: argparse.ArgumentParser)->argparse.ArgumentParser: ...
def parse_service(parser: argparse.ArgumentParser)->argparse.ArgumentParser: ...
def parse_act(parser: argparse.ArgumentParser)->argparse.ArgumentParser: ...
def parse_script(parser: argparse.ArgumentParser)->argparse.ArgumentParser: ...
def apply_act(args: CLI)->CLI: ...


import builtins
import multiprocessing.connection
from abc import abstractmethod
from typing import Any

class ObjectDetectionBase:
    def store_args(self:"ObjectDetectionBase", args:Any=None, unknown_args:Any=None) -> None: ...
    @classmethod
    def print_inferences(cls:type["ObjectDetectionBase"]): ...
    @classmethod
    def get_inference(cls:type["ObjectDetectionBase"], name:builtins.str=None, module:builtins.str=None): ...
    @abstractmethod
    def load_model(self:"ObjectDetectionBase", *args, **kwargs): ...
    @abstractmethod
    def unload_model(self:"ObjectDetectionBase", *args, **kwargs): ...
    @abstractmethod
    def preprocess(self:"ObjectDetectionBase", data: Any, *args, **kwargs): ...
    @abstractmethod
    def postprocess(self:"ObjectDetectionBase", data: Any, *args, **kwargs): ...
    @abstractmethod
    def predict(self:"ObjectDetectionBase", data: Any, *args, **kwargs): ...
    def run(self:"ObjectDetectionBase", data: Any, *args, **kwargs): ...
    def run_async(self: "ObjectDetectionBase", pipe: multiprocessing.connection.Connection): ...
    def run_pipeline(self: "ObjectDetectionBase", pipe: multiprocessing.connection.Connection): ...



import logging

from loguru import logger


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where logged message originated,
        # skipping frames to find the correct function name and line number
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


# Configure logging
logging.basicConfig(handlers=[InterceptHandler()], level=0)



from abc import abstractmethod
from typing import Any

from loguru import logger

from more_than_inference.utils.register import META


class MetaInference(type):
    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        for _, meta in META.registers.items():
            if meta in bases:
                meta.sub_classes.append(new_class)
        return new_class


class ModuleMixin:
    """This Mixin provide operations for find and get subclass of registed MetaInference"""
    @classmethod
    def print_inferences(cls):
        """Print all inferences that registed in MetaInference
        """
        for subclass in cls.sub_classes:
            logger.info("loaded:: "+subclass.__module__ + '.' + subclass.__name__)

    @classmethod
    def get_inference(cls, name: str = None, module: str = None):
        """Get inference from registed
        """
        collects = []
        for subclass in cls.sub_classes:
            if name is None:
                collects.append(subclass)
            else:
                if module not in [None, subclass.__module__]:
                    continue
                if name == subclass.__name__:
                    collects.append(subclass)
        if len(collects) == 0:
            print(f"get none of this inference module=`{module}` name=`{name}`")
            return
        elif len(collects) > 1:
            print(f"get more than one inference in module=`{module}` name=`{name}`, result=[{collects}], "
                  f"would choose `{collects[0]}` as default.")
        return collects[0]


class InferMixin:
    """This Mixin contains load/unload model and preprocess/postprocess/predict data from inference."""
    @abstractmethod
    def load_model(self, *args, **kwargs): raise NotImplementedError
    @abstractmethod
    def unload_model(self, *args, **kwargs): raise NotImplementedError
    @abstractmethod
    def preprocess(self, data: Any, *args, **kwargs): raise NotImplementedError
    @abstractmethod
    def postprocess(self, data: Any, *args, **kwargs): raise NotImplementedError
    @abstractmethod
    def predict(self, data: Any, *args, **kwargs): raise NotImplementedError


class RuntimeMixin:
    def run(self, data: Any, *args, **kwargs):
        data = self.preprocess(data, *args, **kwargs)
        data = self.predict(data, *args, **kwargs)
        data = self.postprocess(data, *args, **kwargs)
        return data


class ArgParserMixin:
    def store_args(self, args=None, unknown_args=None) -> None:
        self.args = args
        self.unknown_args = unknown_args



import asyncio
from typing import Callable

import zmq
import zmq.asyncio
from loguru import logger

from more_than_inference.protos.node_pb2 import Node


class ZeroMQServer:
    def __init__(self, bind_addr: str, node_name: str = None, node_key: str = None, domain: str = None):
        self.bind_addr = bind_addr
        self.node_name = node_name
        self.node_key = node_key
        self.domain = domain
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(zmq.REP)

    async def start(self, callback_func: Callable[[bytes], bytes]):
        self.socket.bind(self.bind_addr)
        while True:
            message = await self.socket.recv()
            await self.socket.send(callback_func(message))


class ZeroMQClient:
    def __init__(self, bind_addr: str = None, node_name: str = None, node_key: str = None, domain: str = None):
        self.bind_addr = bind_addr
        self.node_name = node_name
        self.node_key = node_key
        self.domain = domain
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(zmq.REQ)

    async def send(self, data: bytes, callback_func: Callable[["ZeroMQClient", bytes, bytes], bytes] = None):
        self.socket.connect(self.bind_addr)
        await self.socket.send(data)
        reply = await self.socket.recv()
        if callback_func is not None:
            logger.info(f"CONN TO `{self.bind_addr}` RECV BACK `{reply}`")
            callback_func(self, data, reply)
        self.socket.disconnect(self.bind_addr)


async def main():
    # Server
    def callback_func(inputs: bytes):
        node = Node()
        node.ParseFromString(inputs)
        print(node.Key)
        return b"OK"
    server = ZeroMQServer("ipc:///tmp/zero.sock", "local", "local", "local")
    asyncio.create_task(server.start(callback_func))

    # Give the server some time to start
    # await asyncio.sleep(1)

    # Client
    client = ZeroMQClient("ipc:///tmp/zero.sock", "local", "local", "local")
    node = Node(
        RemoteSocket="tcp://192.168.1.1:5555/",
        LocalSocket="/tmp/node",
        NodeName="ExampleNode",
        Domain="TEST",
        Category="ObjectDetection",
        Key="secretkey"
    )

    def callback_func(*args):
        print(*args)
        return
    await client.send(node.SerializeToString(), callback_func)

if __name__ == "__main__":
    asyncio.run(main())



import os

import setuptools
from setuptools.command.install import install

PACKAGE_NAME = "more_than_inference"


def build_proto(dist_dir, proto_dir):
    os.system(f'protoc --python_out={dist_dir} --proto_path={dist_dir} {proto_dir}/*.proto')


class InstallCommand(install):
    def run(self):
        install.run(self)
        self.compile_protos()

    def compile_protos(self):
        from pathlib import Path
        install_dir = Path(self.install_lib, PACKAGE_NAME)
        proto_dir = install_dir / 'protos'
        build_proto(install_dir.parent, proto_dir)


setuptools.setup(
    packages=setuptools.find_namespace_packages("src"),
    package_dir={"": "src"},
    package_data={PACKAGE_NAME: ["protos/*"]},
    cmdclass={
        'install': InstallCommand,
    },
)



import builtins
from abc import abstractmethod
from typing import Any, Callable, TypeVar, get_args, get_origin

import google.protobuf.message

from more_than_inference.protos.data_pb2 import Data
from more_than_inference.protos.node_pb2 import Node
from more_than_inference.protos.pipeline_pb2 import Pipeline

# ---------------------------------------------------------------------------------------------------------------------
# Serialize Functions
# ---------------------------------------------------------------------------------------------------------------------
_T = TypeVar("_T")
_VT = TypeVar("_VT")
_BUILTINS = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]


class ProtoLoadsMixin:
    def loads(T: _T, data_class: type[google.protobuf.message.Message]) -> Callable[[bytes], _T]:
        def _call_stack(_t: type, _d: Any, self=None):
            if _t in _BUILTINS:
                return _d
            if self is None:
                _i = _t()
            else:
                _i = self
            for field in _t._protected:
                if field not in _t._registed:
                    setattr(_i, field, getattr(_d, field))
                else:
                    registed_type = _t._registed[field]
                    origin_type = get_origin(registed_type)
                    if origin_type is None:
                        setattr(_i, field, _call_stack(registed_type, getattr(_d, field)))
                    elif origin_type == list:
                        item_type = get_args(registed_type)[0]
                        setattr(_i, field, [_call_stack(item_type, __i) for __i in getattr(_d, field)])
                    else:
                        raise ValueError(f"unsupport origin type `{origin_type}` for `{registed_type}`")
            return _i

        def _loads(self: _T, data: bytes) -> _T:
            _inst = data_class()
            _inst.ParseFromString(data)
            return _call_stack(T, _inst, self=self)
        return _loads


class ProtoDumpsMixin:
    def dumps(T: _T, data_class: type[google.protobuf.message.Message]) -> Callable[[_T], bytes]:
        def _call_stack(_t: type, _d: Any, _inst: _VT = None) -> _VT:
            if _t in _BUILTINS:
                return _d
            if _inst is None:
                _inst = _t._data_class()
            for field in _t._protected:
                if not hasattr(_d, field):
                    continue
                if field not in _t._registed:
                    setattr(_inst, field, getattr(_d, field))
                else:
                    registed_type = _t._registed[field]
                    origin_type = get_origin(registed_type)
                    if origin_type is None:
                        _call_stack(registed_type, getattr(_d, field), getattr(_inst, field))
                    elif origin_type == list:
                        item_type = get_args(registed_type)[0]
                        for __i in getattr(_d, field):
                            getattr(_inst, field).append(_call_stack(item_type, __i))
                    else:
                        raise ValueError(f"unsupport origin type `{origin_type}` for `{registed_type}`")
            return _inst

        def _dumps(self: _T) -> bytes:
            _inst = _call_stack(T, self, data_class())
            return _inst.SerializeToString()
        return _dumps

# TODO: Json implymention


class JsonLoadsMixin:
    def loads():
        ...


class JsonDumpsMixin:
    def dumps():
        ...


# ---------------------------------------------------------------------------------------------------------------------
# Serializers
# ---------------------------------------------------------------------------------------------------------------------
class BaseSerializer:
    @abstractmethod
    def loads(data: bytes) -> Any: raise NotImplementedError
    @abstractmethod
    def dumps(data: Any) -> bytes: raise NotImplementedError


class MetaSerializer(type):
    def __new__(cls: type, name: str, bases: tuple, attrs: dict):
        new_class = super().__new__(cls, name, bases, attrs)
        _data_class = attrs.get('_data_class', None)  # `_data_class` defined the main protocol class in use
        _protected = attrs.get('_protected', None)  # `_protected` defined the field to serialize
        _registed = attrs.get('_registed', None)  # `_registed` defined the field already registed
        if _data_class is not None:
            setattr(new_class, 'loads', cls.loads(new_class, _data_class))
            setattr(new_class, 'dumps', cls.dumps(new_class, _data_class))
        # set default value
        if _protected is None:
            setattr(new_class, '_protected', attrs['__annotations__'].keys())
        if _registed is None:
            setattr(new_class, '_registed', {})
        return new_class


class ProtoSerializer(ProtoLoadsMixin, ProtoDumpsMixin, MetaSerializer):
    """This Class is a MetaClass, provide the loads/dumps function to childs, Backend support proto.
    To enable using loads/dumps function, define class var `_data_class`. Define class var `_protected`
    should give those specified key data to transfer.
    """
    ...


# ---------------------------------------------------------------------------------------------------------------------
# Data Classes
# ---------------------------------------------------------------------------------------------------------------------
class TypedData(BaseSerializer, metaclass=ProtoSerializer):
    _data_class = Data

    Body: builtins.bytes
    Serializer: builtins.str
    Seq: builtins.int
    Ended: builtins.bool


class TypedNode(BaseSerializer, metaclass=ProtoSerializer):
    _data_class = Node

    RemoteSocket: builtins.str
    LocalSocket: builtins.str
    NodeName: builtins.str
    Domain: builtins.str
    Category: builtins.str
    Key: builtins.str


class TypedPipeline(BaseSerializer, metaclass=ProtoSerializer):
    _data_class = Pipeline
    _registed = {"StackList": list[TypedNode], "Data": TypedData}

    Key: builtins.str
    StackList: list[TypedNode]
    Data: TypedData


if __name__ == "__main__":
    data = TypedData()
    data.Body = b''
    data.Serializer = "hello"
    data.Seq = 0
    data.Ended = False
    bdata = data.dumps()
    print(bdata)

    cdata = TypedData()
    cdata.loads(bdata)
    print(cdata.Serializer)
    print(get_args(list[TypedNode]))
    print(get_origin(TypedNode))

    tp = TypedPipeline()
    tp.Data = TypedData()
    tp.Data.Body = b'\n\x04KKKK\x12\n\x1a\x03GGG2\x03SST\x1a\x05\x12\x03DIM'
    tp.Data.Ended = False
    tp.Data.Seq = 0
    tp.Data.Serializer = "DIM"
    tp.StackList = []
    node = TypedNode()
    node.Category = "SST"
    node.LocalSocket = "GGG"
    tp.StackList.append(node)
    tp.Key = "KKKK"

    ss = tp.dumps()
    print(ss)

    ttp = TypedPipeline()
    ttp.loads(ss)
    print(ttp.StackList[0].LocalSocket)
    print(ttp.Data.Serializer)
